#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. currentmodule:: normanpg.cli
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This is the entry point for the command-line interface (CLI) application.  It
can be used as a handy facility for running the task from a command line.

.. note::

    To learn more about Click visit the
    `project website <http://click.pocoo.org/5/>`_.  There is also a very
    helpful `tutorial video <https://www.youtube.com/watch?v=kNke39OZ2k0>`_.

    To learn more about running Luigi, visit the Luigi project's
    `Read-The-Docs <http://luigi.readthedocs.io/en/stable/>`_ page.
"""
import logging
import click
from .__init__ import __version__

LOGGING_LEVELS = {
    0: logging.NOTSET,
    1: logging.ERROR,
    2: logging.WARN,
    3: logging.INFO,
    4: logging.DEBUG
}  #: a mapping of `verbose` option counts to logging levels


class Info(object):
    """
    This is an information object that can be used to pass data between CLI functions.
    """
    def __init__(self):  # Note that this object must have an empty constructor.
        self.verbose: int = 0


# pass_info is a decorator for functions that pass 'Info' objects.
#: pylint: disable=invalid-name
pass_info = click.make_pass_decorator(
    Info,
    ensure=True
)


# Change the options to below to suit the actual options for your task (or
# tasks).
@click.group()
@click.option('--verbose', '-v', count=True, help="Enable verbose output.")
@pass_info
def cli(info: Info,
        verbose: int):
    """
    Run normanpg.
    """
    # Use the verbosity count to determine the logging level...
    if verbose > 0:
        logging.basicConfig(
            level=LOGGING_LEVELS[verbose]
            if verbose in LOGGING_LEVELS
            else logging.DEBUG
        )
        click.echo(
            click.style(
                f'Verbose logging is enabled. '
                f'(LEVEL={logging.getLogger().getEffectiveLevel()})',
                fg='yellow'
            )
        )
    info.verbose = verbose

@cli.command()
@pass_info
def hello(_: Info):
    """
    Say 'hello' to the nice people.
    """
    click.echo(f"normanpg says 'hello'")


@cli.command()
def version():
    """
    Get the library version.
    """
    click.echo(click.style(f'{__version__}', bold=True))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. currentmodule:: normanpg.version
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This module contains project version information.
"""

__version__ = '0.0.4'  #: the working version
__release__ = '0.0.4'  #: the release version



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 6/29/19
"""
.. currentmodule:: normanpg.errors
.. moduleauthor:: Pat Daburu <pat@daburu.net>

Sometimes things go wrong.
"""


class NormanPgException(Exception):
    """
    This is the base exception for other exceptions defined in this library.
    """
    def __init__(self, message: str, inner: Exception = None):
        """

        :param message: the exception message
        :param inner: the exception that caused this exception
        """
        super().__init__(message)
        self._message = message
        self._inner = inner

    @property
    def message(self) -> str:
        """
        Get the exception message.
        """
        return self._message

    @property
    def inner(self) -> Exception or None:
        """
        Get the exception that caused this exception.
        """
        return self._inner



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. currentmodule:: normanpg
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This is a set of modest utilities that may be helpful when talking to
PostgreSQL.
"""
from .pg import connect, execute, execute_rows, execute_scalar
from .version import __version__, __release__



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created on 9/22/19 by Pat Blair
"""
.. currentmodule:: pg
.. moduleauthor:: Pat Blair <pblair@geo-comm.com>

This module needs a description.
"""
import inspect
import logging
from typing import Any, Iterable, Union
from urllib.parse import urlparse, ParseResult
import psycopg2.extras
import psycopg2.sql
import psycopg2.extensions


__logger__ = logging.getLogger(__name__)  #: the module logger

DEFAULT_PG_PORT: int = 5432  #: the default Postgres database port


def connect(
        url: str,
        dbname: str = None,
        autocommit: bool = False
) -> psycopg2.extensions.connection:
    """
    Get a connection to a Postgres database instance.

    :param url: the instance URL
    :param dbname: the target database name
    :param autocommit: Set the `autocommit` flag on the connection?
    :return: a psycopg2 connection

    .. note::

        If the caller does not provide the `dbname` parameter the function
        creates a connection to the database specified in the URL.
    """
    # Parse the URL.  (We'll need the pieces to construct a connection
    # string.)
    dbp: ParseResult = urlparse(url)
    # Create a dictionary to hold the arguments for the connection.  (We'll
    # unpack it later.)
    cnx_opt = {
        k: v for k, v in
        {
            'host': dbp.hostname,
            'port': int(dbp.port) if dbp.port is not None else DEFAULT_PG_PORT,
            'database': dbname if dbname is not None else dbp.path[1:],
            'user': dbp.username,
            'password': dbp.password
        }.items() if v is not None
    }
    cnx = psycopg2.connect(**cnx_opt)
    # If the caller requested that the 'autocommit' flag be set...
    if autocommit:
        # ...do that now.
        cnx.autocommit = True
    return cnx


def _execute_scalar(
        cnx: psycopg2.extensions.connection,
        query: psycopg2.sql.Composed,
        caller: str
) -> Any:
    """
    This is a helper function for :py:func:`execute_scalar` that executes a
    query on an open cursor.

    :param cnx: an open connection or database connection string
    :param query: the query
    :param caller: identifies the call stack location
    """
    with cnx.cursor() as crs:
        # Log the query.
        __logger__.debug(f'[{caller}] {query.as_string(crs)}')
        # Execute!
        try:
            crs.execute(query)
        except SyntaxError:
            logging.exception(query.as_string(crs))
            raise
        # Get the first column from the first result.
        return crs.fetchone()[0]


def execute_scalar(
        cnx: Union[str, psycopg2.extensions.connection],
        query: Union[str, psycopg2.sql.Composed],
        caller: str = None
) -> Any or None:
    """
    Execute a query that returns a single, scalar result.

    :param cnx: an open psycopg2 connection or the database URL
    :param query: the `psycopg2` composed query
    :param caller: identifies the caller (for diagnostics)
    :return: the scalar string result (or `None` if the query returns no
        result)
    """
    # Get the name of the calling function so we can include it in the logging
    # statement.
    caller = caller if caller else inspect.stack()[1][3]
    # Make sure the query is `Composed`.
    _query = (
        psycopg2.sql.SQL(query).format()
        if isinstance(query, str)
        else query
    )
    # If the caller passed us a connection string...
    if isinstance(cnx, str):
        # ...get a connection and use the helper method to execute the query.
        with connect(url=cnx) as _cnx:
            return _execute_scalar(cnx=_cnx, query=_query, caller=caller)
    # It looks as though we were given an open connection, so execute the
    # query on it.
    return _execute_scalar(cnx=cnx, query=_query, caller=caller)


def _execute_rows(
        cnx: psycopg2.extensions.connection,
        query: psycopg2.sql.Composed,
        caller: str
) -> Iterable[psycopg2.extras.DictRow]:
    """
    This is a helper function for :py:func:`execute_rows` that executes a
    query on an open cursor.

    :param cnx: an open connection or database connection string
    :param query: the query
    :param caller: identifies the call stack location
    :return: an iteration of `DictRow` instances representing the rows
    """
    with cnx.cursor(cursor_factory=psycopg2.extras.DictCursor) as crs:
        # Log the query.
        __logger__.debug(f'[{caller}] {query.as_string(crs)}')
        # Execute!
        try:
            crs.execute(query)
        except SyntaxError:
            logging.exception(query.as_string(crs))
            raise
        # Fetch the rows and yield them to the caller.
        for row in crs:
            yield row


def execute_rows(
        cnx: Union[str, psycopg2.extensions.connection],
        query: Union[str, psycopg2.sql.Composed],
        caller: str = None
) -> Iterable[psycopg2.extras.DictRow]:
    """
    Execute a query that returns an iteration of rows.

    :param cnx: an open connection or database connection string
    :param query: the `psycopg2` composed query
    :param caller: identifies the caller (for diagnostics)
    :return: an iteration of `DictRow` instances representing the row
    """
    # Get the name of the calling function so we can include it in the logging
    # statement.
    caller = caller if caller else inspect.stack()[1][3]
    # Make sure the query is `Composed`.
    _query = (
        psycopg2.sql.SQL(query).format()
        if isinstance(query, str)
        else query
    )
    # If the caller passed us a connection string...
    if isinstance(cnx, str):
        # ...get a connection and use the helper method to execute the query.
        with connect(url=cnx) as _cnx:
            for row in _execute_rows(cnx=_cnx, query=_query, caller=caller):
                yield row
    # It looks as though we were given an open connection, so execute the
    # query on it.
    for row in _execute_rows(cnx=cnx, query=_query, caller=caller):
        yield row


def _execute(
        cnx: psycopg2.extensions.connection,
        query: psycopg2.sql.Composed,
        caller: str
):
    """
    This is a helper function for :py:func:`execute` that executes a
    query on an open cursor.

    :param cnx: an open connection or database connection string
    :param query: the query
    :param caller: identifies the call stack location
    """
    with cnx.cursor() as crs:
        # Log the query.
        __logger__.debug(f'[{caller}] {query.as_string(crs)}')
        # Execute!
        try:
            crs.execute(query)
        except SyntaxError:
            logging.exception(query.as_string(crs))
            raise


def execute(
        cnx: Union[str, psycopg2.extensions.connection],
        query: Union[str, psycopg2.sql.Composed],
        caller: str = None
):
    """
    Execute a query that returns no result.

    :param cnx: an open connection or database connection string
    :param query: the `psycopg2` composed query
    :param caller: identifies the caller (for diagnostics)

    .. seealso::

        * :py:func:`execute_scalar`
        * :py:func:`execute_rows`
    """
    # Get the name of the calling function so we can include it in the logging
    # statement.
    caller = caller if caller else inspect.stack()[1][3]
    # Make sure the query is `Composed`.
    _query = (
        psycopg2.sql.SQL(query).format()
        if isinstance(query, str)
        else query
    )
    # If the caller passed us a connection string...
    if isinstance(cnx, str):
        # ...get a connection and use the helper method to execute the query.
        with connect(url=cnx) as _cnx:
            _execute(cnx=_cnx, query=_query, caller=caller)
    # It looks as though we were given an open connection, so execute the
    # query on it.
    _execute(cnx=cnx, query=_query, caller=caller)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created on 9/22/19 by pat
"""
.. currentmodule:: normanpg.geometry
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This module contains conveniences for working with geometries.
"""
from shapely.geometry.base import BaseGeometry
from shapely import wkb


def shape(obj: str) -> BaseGeometry:
    """
    Convert a geometry from Postgres into a
    `Shapely <https://shapely.readthedocs.io/en/stable/manual.html#geometric-objects>`_
    geometry.

    :param obj: the raw geometry retrieved from Postgres (a WKB hex string)
    :return: the `Shapely` geometry
    """
    return wkb.loads(obj, hex=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. currentmodule:: setup.py
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This file is used to create the package we'll publish to PyPI.
"""

import importlib.util
import os
from pathlib import Path
from setuptools import setup, find_packages, Command
from codecs import open  # Use a consistent encoding.
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Get the base version from the library.  (We'll find it in the `version.py`
# file in the src directory, but we'll bypass actually loading up the library.)
vspec = importlib.util.spec_from_file_location(
  "version",
  str(Path(__file__).resolve().parent / 'normanpg' / "version.py")
)
vmod = importlib.util.module_from_spec(vspec)
vspec.loader.exec_module(vmod)
version = getattr(vmod, '__version__')

# If the environment has a build number set...
if os.getenv('buildnum') is not None:
    # ...append it to the version.
    version = "{version}.{buildnum}".format(
        version=version,
        buildnum=os.getenv('buildnum')
    )

setup(
    name='normanpg',
    description="This is a set of modest utilities that may be helpful when talking to PostgreSQL.",
    long_description=long_description,
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    version=version,
    install_requires=[
        # Include your dependencies here.
        # Here are a couple of examples...
        # 'numpy>=1.13.3,<2',
        # 'measurement>=1.8.0,<2'
        'click>=7.0,<8',
        'psycopg2-binary>=2.7.4,<3',
        'shapely[vectorized]'
    ],
    entry_points="""
    [console_scripts]
    normanpg=normanpg.cli:cli
    """,
    python_requires=">=0.0.1",
    license='MIT',
    author='Pat Daburu',
    author_email='pat@daburu.net',
    # Use the URL to the github repo.
    url='https://github.com/patdaburu/normanpg',
    download_url=(
        f'https://github.com/patdaburu/'
        f'normanpg/archive/{version}.tar.gz'
    ),
    keywords=[
        # Add package keywords here.
    ],
    # See https://PyPI.python.org/PyPI?%3Aaction=list_classifiers
    classifiers=[
      # How mature is this project? Common values are
      #   3 - Alpha
      #   4 - Beta
      #   5 - Production/Stable
      'Development Status :: 3 - Alpha',

      # Indicate who your project is intended for.
      'Intended Audience :: Developers',
      'Topic :: Software Development :: Libraries',

      # Pick your license.  (It should match "license" above.)
      'License :: OSI Approved :: MIT License',

      # Specify the Python versions you support here. In particular, ensure
      # that you indicate whether you support Python 2, Python 3 or both.
      'Programming Language :: Python :: 3.6',
    ],
    include_package_data=True
)
